﻿
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using static System.Collections.Specialized.BitVector32;
using Mono.CSharp;

namespace Preset_Updater
{

    public class PresetUpdater : MVRScript
    {
        UIDynamicButton _processVAPPresetsButton;
        UIDynamicButton _retakeScreenShots;
        UIDynamicPopup _presetTypePopup;
        UIDynamicSlider _delaySlider;
        
        public PresetManager _pm;

        private const string _cameraSavePath = "Custom/PluginData/bill_prime/PresetUpdater/";
        private const int _jsonstringLimit = 15000;
        private const string _atomPersonPath = "Custom/Atom/Person/";
        private const string _addonPackagesPath = "AddonPackages";
        private const string _filePrefs = "AddonPackagesFilePrefs/";
        private const string _clothing = "Clothing";
        private const string _hair = "Hair";
        private const string _pose = "Pose";
        private const string _skin = "Skin";
        private const string _appearance = "Appearance";
        private string _presetName;
        private string _presetPath;
        private string _sceneVar;
        private string _varFileName;
        private Dictionary<string, JSONStorableBool> _enablePresetDic = new Dictionary<string, JSONStorableBool>();
        private Dictionary<string, UnityAction> _setPath = new Dictionary<string, UnityAction>();
        private Dictionary<string, string> _presetSystemPath = new Dictionary<string, string>();
        private Dictionary<string, string> _presetTypeID = new Dictionary<string, string>();
        private Dictionary<string, JSONStorableString> _savedPathDic = new Dictionary<string, JSONStorableString>();
        private Dictionary<string, string> _presetPathDic = new Dictionary<string, string>();
        private Dictionary<string, List<string>> _presetListDic = new Dictionary<string, List<string>>();
        private Dictionary<string, List<string>> _listDic = new Dictionary<string, List<string>>();
        private Dictionary<string, JSONStorableString> _filesInFolder = new Dictionary<string, JSONStorableString>();
        private Dictionary<string, UIDynamicButton> _setPresetPathButton = new Dictionary<string, UIDynamicButton>();
        private Dictionary<string, UIDynamicToggle> _enablePresetToggle = new Dictionary<string, UIDynamicToggle>();
        private List<string> _presetList = new List<string>();

        private UIDynamicButton _setPresetPathButtonSelected;
        private JSONStorableStringChooser _presetTypeChooser;
        private string _selectedPresetType;
        private JSONStorableFloat _screenshotDelay;

        DAZCharacterSelector _dazCharacterSelector;

        CamPositions _camPositions = new CamPositions();
        private Dictionary<string, Color> _camHelpColorDic = new Dictionary<string, Color>();

        private bool _boolcontinue = false;
        private bool _stopRequested = false;
        private int _presetCount = 0;
        private bool _isProcessing = false;
        private bool _needToSaveCams = true;
        private bool _waitingOnScreenShot = false;

        Transform _headTransform;
        InfoBox _infoBox;

        private readonly string[] _presets = new string[]
        {
            "Appearance",
            "Clothing",
            "Hair",
            "Skin",
            "Pose"
        };

        public override void Init()
        {
            var sceneDir = SuperController.singleton.currentLoadDir;
            if (sceneDir.Contains(":"))
            {
                var sceneDirSplit = sceneDir.Split(':');
                _sceneVar = sceneDirSplit[0];
            }
            else
            {
                _sceneVar = "";
            }

            JSONStorableAction SetClothingPresetPathAct = new JSONStorableAction("Set " + _clothing + " Presets Path", SetClothingPathAct);
            JSONStorableAction SetHairPresetPathAct = new JSONStorableAction("Set " + _hair + " Presets Path", SetHairPathAct);
            JSONStorableAction SetPosePresetPathAct = new JSONStorableAction("Set " + _pose + " Presets Path", SetPosePathAct);
            JSONStorableAction SetSkinPresetPathAct = new JSONStorableAction("Set " + _skin + " Presets Path", SetSkinPathAct);
            JSONStorableAction SetAppearancePresetPathAct = new JSONStorableAction("Set " + _appearance + " Presets Path", SetAppearancePathAct);
            _setPath[_clothing] = SetClothingPath;
            _setPath[_hair] = SetHairPath;
            _setPath[_pose] = SetPosePath;
            _setPath[_skin] = SetSkinPath;
            _setPath[_appearance] = SetAppearancePath;
            _savedPathDic[_clothing] = new JSONStorableString("Clothing Presets Path", "");
            _savedPathDic[_hair] = new JSONStorableString("Hair Presets Path", "");
            _savedPathDic[_pose] = new JSONStorableString("Pose Preset Path", "");
            _savedPathDic[_skin] = new JSONStorableString("Skin Preset Path", "");
            _savedPathDic[_appearance] = new JSONStorableString("Appearance Preset Path", "");
            _camHelpColorDic[_clothing] = Color.yellow;
            _camHelpColorDic[_hair] = Color.green;
            _camHelpColorDic[_pose] = Color.black;
            _camHelpColorDic[_skin] = Color.red;
            _camHelpColorDic[_appearance] = Color.cyan;

            //create folders
            if (!FileManagerSecure.DirectoryExists(_cameraSavePath))
            {
                FileManagerSecure.CreateDirectory(_cameraSavePath);
            }

            if (base.containingAtom.mainController == null)
            {
                SuperController.LogError("Please add this plugin to a PERSON atom.");
                return;
            }

            _headTransform = containingAtom.rigidbodies.Where(x => x.name == "head").FirstOrDefault().transform;
            _dazCharacterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();

            _camPositions.StoreCamPosition("HeadPosition", _headTransform);

            JSONStorableString uiInfoBox = new JSONStorableString("uiSetupTimer", "");
            _infoBox = new InfoBox(uiInfoBox);

            UIDynamicTextField t = CreateTextField(uiInfoBox);
            RectTransform rt = t.backgroundImage.GetComponent<RectTransform>();
            Vector2 v = rt.sizeDelta;
            rt.sizeDelta = new Vector2(500, v.y);
            rt.transform.localPosition += new Vector3(250, 0, 0);
            UIDynamic spacerRight = CreateSpacer(true);
            spacerRight.height = 215;

            //StartCoroutine(ScanSetup());
            _selectedPresetType = _clothing;
            _presetTypeChooser = new JSONStorableStringChooser("Preset Type", new List<string>(_presets), "Clothing", "Preset Type", UpdateSelectedPresetType);
            RegisterStringChooser(_presetTypeChooser);

            _screenshotDelay = new JSONStorableFloat("Delay Seconds\n(Allow physics to settle)",1f,0f,2f,constrain: true);

            CreateUI();

            foreach (var item in _presets)
            {
                _presetSystemPath[item] = _atomPersonPath + item;
                _listDic[item] = new List<string>();
                _enablePresetDic[item] = new JSONStorableBool("Enable " + item + " Presets", false);
                _filesInFolder[item] = new JSONStorableString("", "");
                _presetTypeID[item] = item + "Presets";

                //MakeUi(item, _setPath[item]);
            }

            RegisterAction(SetClothingPresetPathAct);
            RegisterAction(SetHairPresetPathAct);
            RegisterAction(SetPosePresetPathAct);
            RegisterAction(SetSkinPresetPathAct);
            RegisterAction(SetAppearancePresetPathAct);

            //RegisterBool(_enablePresetDic[_clothing]);
            //RegisterBool(_enablePresetDic[_hair]);
            //RegisterBool(_enablePresetDic[_pose]);
            //RegisterBool(_enablePresetDic[_skin]);
            //RegisterBool(_enablePresetDic[_appearance]);

            RegisterString(_savedPathDic[_clothing]);
            RegisterString(_savedPathDic[_hair]);
            RegisterString(_savedPathDic[_pose]);
            RegisterString(_savedPathDic[_skin]);
            RegisterString(_savedPathDic[_appearance]);

        }

        private void UpdateSelectedPresetType(JSONStorableStringChooser jsonChooser)
        {
            _selectedPresetType = jsonChooser.val;
            CheckValidPath();
        }

        private void CheckValidPath()
        {
            if (_savedPathDic[_selectedPresetType].val == "")
            {
                _processVAPPresetsButton.button.enabled = false;
                _processVAPPresetsButton.buttonColor = Color.red;
                _setPresetPathButtonSelected.buttonColor = Color.red;
                _infoBox.Status = "Set folder path to update Presets.";
                _infoBox.PresetPath = _savedPathDic[_selectedPresetType].val.Replace("\\", "/"); ;
                _infoBox.VapCount = 0;
                _infoBox.Message = "";                
            }
            else
            {
                _processVAPPresetsButton.button.enabled = true;
                _processVAPPresetsButton.buttonColor = Color.cyan;
                _setPresetPathButtonSelected.buttonColor = Color.green;
                _infoBox.Status = "Ready to start Processing.";
                _infoBox.PresetPath = _savedPathDic[_selectedPresetType].val.Replace("\\", "/"); ;
                _infoBox.VapCount = _presetListDic[_selectedPresetType].Count();
                _infoBox.Message = "";
            }
}

        public void MakeUi(string presetName, UnityAction action)
        {
            _setPresetPathButton[presetName] = CreateButton("Set " + presetName + " Presets Path");
            _setPresetPathButton[presetName].button.onClick.AddListener(action);

            _enablePresetToggle[presetName] = CreateToggle(_enablePresetDic[presetName]);
        }

        public void SetSelectedPath()
        {
            _setPath[_selectedPresetType].Invoke();
        }

        public void SetClothingPath()
        {
            var presetName = _clothing;
            _presetName = presetName;
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_presetSystemPath[presetName]);
            _presetPath = _presetSystemPath[presetName];
            SuperController.singleton.GetDirectoryPathDialog(LoadPresetsList, _presetPath, shortcuts, false);
        }

        public void SetClothingPathAct()
        {
            var presetName = _clothing;
            _presetPath = _presetSystemPath[presetName];
            _presetName = presetName;
            LoadPresetsList(_savedPathDic[presetName].val);
            if (_savedPathDic[presetName].val == "")
            {
                _presetPathDic[presetName] = "";
                _presetListDic[presetName] = new List<string>();
            }
            ChangePath(presetName);
        }
        public void SetHairPath()
        {
            var presetName = _hair;
            _presetName = presetName;
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_presetSystemPath[presetName]);
            _presetPath = _presetSystemPath[presetName];
            SuperController.singleton.GetDirectoryPathDialog(LoadPresetsList, _presetPath, shortcuts, false);
        }

        public void SetHairPathAct()
        {
            var presetName = _hair;
            _presetPath = _presetSystemPath[presetName];
            _presetName = presetName;
            LoadPresetsList(_savedPathDic[presetName].val);
            if (_savedPathDic[presetName].val == "")
            {
                _presetPathDic[presetName] = "";
                _presetListDic[presetName] = new List<string>();
                _enablePresetDic[presetName].val = false;
            }
            ChangePath(presetName);
        }
        public void SetPosePath()
        {
            var presetName = _pose;
            _presetName = presetName;
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_presetSystemPath[presetName]);
            _presetPath = _presetSystemPath[presetName];
            SuperController.singleton.GetDirectoryPathDialog(LoadPresetsList, _presetPath, shortcuts, false);
        }

        public void SetPosePathAct()
        {
            var presetName = _pose;
            _presetPath = _presetSystemPath[presetName];
            _presetName = presetName;
            LoadPresetsList(_savedPathDic[presetName].val);
            if (_savedPathDic[presetName].val == "")
            {
                _presetPathDic[presetName] = "";
                _presetListDic[presetName] = new List<string>();
                _enablePresetDic[presetName].val = false;
            }
            ChangePath(presetName);
        }

        public void SetSkinPath()
        {
            var presetName = _skin;
            _presetName = presetName;
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_presetSystemPath[presetName]);
            _presetPath = _presetSystemPath[presetName];
            SuperController.singleton.GetDirectoryPathDialog(LoadPresetsList, _presetPath, shortcuts, false);
        }

        public void SetSkinPathAct()
        {
            var presetName = _skin;
            _presetPath = _presetSystemPath[presetName];
            _presetName = presetName;
            LoadPresetsList(_savedPathDic[presetName].val);
            if (_savedPathDic[presetName].val == "")
            {
                _presetPathDic[presetName] = "";
                _presetListDic[presetName] = new List<string>();
                _enablePresetDic[presetName].val = false;
            }
            ChangePath(presetName);
        }

        public void SetAppearancePath()
        {
            var presetName = _appearance;
            _presetName = presetName;
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(_presetSystemPath[presetName]);
            _presetPath = _presetSystemPath[presetName];
            SuperController.singleton.GetDirectoryPathDialog(LoadPresetsList, _presetPath, shortcuts, false);
        }

        public void SetAppearancePathAct()
        {
            var presetName = _appearance;
            _presetPath = _presetSystemPath[presetName];
            _presetName = presetName;
            LoadPresetsList(_savedPathDic[presetName].val);
            if (_savedPathDic[presetName].val == "")
            {
                _presetPathDic[presetName] = "";
                _presetListDic[presetName] = new List<string>();
                _enablePresetDic[presetName].val = false;
            }
            ChangePath(presetName);
        }

        public void ChangePath(string presetName)
        {
            _filesInFolder[presetName].val = _presetPathDic[presetName];
            _listDic[presetName] = new List<string>(_presetListDic[presetName]);
        }

        public void LoadPresetsList(string path)
        {
            if (path.Contains(":")) 
            {
                //no vars!
                _savedPathDic[_presetName].val = "";

                _processVAPPresetsButton.button.enabled = false;
                _processVAPPresetsButton.buttonColor = Color.red;
                _setPresetPathButtonSelected.buttonColor = Color.red;
                _infoBox.Status = "Set folder path to update Presets.";
                _infoBox.PresetPath = _savedPathDic[_presetName].val.Replace("\\","/");
                _infoBox.VapCount = 0;
                _infoBox.Message = "Error: Do not select a folder in a var file, local files only.";

                _filesInFolder[_presetName] = new JSONStorableString("", "");
                _listDic[_presetName] = new List<string>();
                _presetPathDic[_presetName] = "";
                _presetListDic[_presetName] = new List<string>();
                _enablePresetDic[_presetName].val = false;

                return;
            }
            if (path != null && path != "")
            {
                _savedPathDic[_presetName].val = path;
                path = path.Replace("\\", "/");
                path = MakeFilename(path);

                if (path.Substring(path.Length - 1) == "/")
                {
                    path = path.Remove(path.Length - 1, 1);
                }

                if (!path.Contains(":") && _sceneVar != "")
                {
                    var tempPath = _sceneVar + ":/" + path;
                    if (!FileManagerSecure.DirectoryExists(path))
                    {
                        path = tempPath;
                    }
                }

                _varFileName = "";
                if (path.Contains(":"))
                {
                    var pathSplit = path.Split(':');
                    _varFileName = pathSplit[0];
                }

                string[] presetFiles;
                string presetFileList = "";

                presetFiles = FileManagerSecure.GetFiles(path, "Preset_*.vap");
                Array.Sort(presetFiles);
                _presetList = new List<string>();
                foreach (var item in presetFiles)
                {
                    string tempFileName = item.Replace("\\", "/");
                    if (tempFileName.Contains(":"))
                    {
                        var itemSplit = tempFileName.Split(':');
                        tempFileName = _varFileName + ":" + itemSplit[1].Remove(0, _presetSystemPath[_presetName].Length + 2);
                    }
                    else
                    {
                        tempFileName = tempFileName.Remove(0, _presetSystemPath[_presetName].Length + 1);
                    }

                    tempFileName = tempFileName.Remove(tempFileName.LastIndexOf("Preset_"), 7);
                    tempFileName = tempFileName.Remove(tempFileName.LastIndexOf(".vap"), 4);
                    _presetList.Add(tempFileName);
                    if (presetFileList.Length < _jsonstringLimit)
                    {
                        presetFileList = presetFileList + tempFileName + "\n";
                    }
                }

                _presetListDic[_presetName] = new List<string>(_presetList);
                _presetPathDic[_presetName] = presetFileList;

                _filesInFolder[_presetName].val = _presetPathDic[_presetName];
                _listDic[_presetName] = new List<string>(_presetListDic[_presetName]);

                _processVAPPresetsButton.button.enabled = true;
                _processVAPPresetsButton.buttonColor = Color.cyan;
                _setPresetPathButtonSelected.buttonColor = Color.green;
                _infoBox.Status = "Ready to start Processing.";
                _infoBox.PresetPath = _savedPathDic[_selectedPresetType].val.Replace("\\", "/"); ;
                _infoBox.VapCount = _presetListDic[_selectedPresetType].Count();
                _infoBox.Message = "";
            }
            else
            {
                _processVAPPresetsButton.button.enabled = false;
                _processVAPPresetsButton.buttonColor = Color.red;
                _setPresetPathButtonSelected.buttonColor = Color.red;
                _infoBox.Status = "Set folder path to update Presets.";
                _infoBox.PresetPath = "";
                _infoBox.VapCount = _presetListDic[_selectedPresetType].Count();
                _infoBox.Message = "";
            }
        }

        public string MakeFilename(string path)
        {
            if (path != null && path != "")
            {
                path = path.Replace("\\", "/");
                if (!path.Contains(":"))
                {
                    if (path.Substring(0, 1) == "/")
                    {
                        path = path.Remove(0, 1);
                    }

                    if (!path.Contains(_presetPath))
                    {
                        path = _presetPath + "/" + path;
                    }
                }
                else
                {
                    if (path.Contains(".var:"))
                    {
                        path = path.Replace(".var:", ":");
                    }

                    if (!path.Contains(":/"))
                    {
                        path = path.Replace(":", ":/");
                    }

                    string[] div = { ":/" };
                    var pathSplit = path.Split(div, StringSplitOptions.None);

                    if (!pathSplit[1].Contains(_presetSystemPath[_presetName]))
                    {
                        pathSplit[1] = (_presetSystemPath[_presetName] + "/" + pathSplit[1]);
                    }

                    var varPathSplit = pathSplit[0].Split('/');
                    var varNameFull = varPathSplit[varPathSplit.Length - 1];
                    var varNameSplit = varNameFull.Split('.');
                    var varName = varNameSplit[0] + "." + varNameSplit[1];
                    var varExist = FileManagerSecure.FileExists(pathSplit[0] + ".var");
                    if (varExist)
                    {
                        path = varNameFull + ":/" + pathSplit[1];
                    }
                    else
                    {
                        varExist = FileManagerSecure.FileExists(_addonPackagesPath + "/" + varNameFull + ".var");
                        if (varExist)
                        {
                            path = varNameFull + ":/" + pathSplit[1];
                        }
                        else
                        {
                            int i = 100;
                            bool varExistFull;
                            bool varExistKy;
                            do
                            {
                                i--;
                                varExistFull = FileManagerSecure.FileExists(pathSplit[0].Remove(pathSplit[0].LastIndexOf(".")) + "." + i.ToString() + ".var");
                                varExist = FileManagerSecure.FileExists(_addonPackagesPath + "/" + varName + "." + i.ToString() + ".var");
                                varExistKy = FileManagerSecure.FileExists(_addonPackagesPath + "/" + _addonPackagesPath + "/" + varName + "." + i.ToString() + ".var");
                            }
                            while (!varExistFull && !varExist && !varExistKy && i > 0);

                            if (i != 0)
                            {
                                path = varName + "." + i.ToString() + ":/" + pathSplit[1];
                            }
                            else
                            {
                                ArrayList allFiles = new ArrayList();
                                SearchFile(_addonPackagesPath, varName + ".*.var", ref allFiles);
                                var searchedFiles = (string[])allFiles.ToArray(typeof(string));
                                List<string> varFiles = new List<string>();

                                if (searchedFiles != null)
                                {
                                    foreach (var item in searchedFiles)
                                    {
                                        var searchedFilesSplit = item.Replace("\\", "/").Split('/');
                                        var searchedVarName = searchedFilesSplit[searchedFilesSplit.Length - 1];
                                        searchedVarName = searchedVarName.Remove(searchedVarName.LastIndexOf(".var"));
                                        varFiles.Add(searchedVarName);
                                    }

                                    if (varFiles.Contains(varNameFull))
                                    {
                                        varName = varNameFull;
                                    }
                                    else
                                    {
                                        int latest = 1;
                                        foreach (var item in varFiles)
                                        {
                                            var itemSplit = item.Split('.');
                                            if (latest <= int.Parse(itemSplit[2]))
                                            {
                                                latest = int.Parse(itemSplit[2]);
                                            }
                                        }
                                        varName = varNameFull.Remove(varNameFull.LastIndexOf(".")) + "." + latest.ToString();
                                    }
                                    path = varName + ":/" + pathSplit[1];
                                }
                                else
                                {
                                    path = "";
                                }
                            }
                        }
                    }
                }
                return path;
            }
            return "";
        }

        public static void SearchFile(string folder, string filter, ref ArrayList allFiles)
        {
            string[] filses = FileManagerSecure.GetFiles(folder, filter);
            allFiles.AddRange(filses);

            string[] subFolders = FileManagerSecure.GetDirectories(folder);
            foreach (string subF in subFolders)
            {
                SearchFile(subF, filter, ref allFiles);
            }
        }

        void EnableUIButtons(bool state)
        {
            _presetTypeChooser.storable.enabled = state;
            _setPresetPathButtonSelected.button.enabled = state;
            _retakeScreenShots.button.enabled = state;
            _delaySlider.slider.enabled = state;
            _processVAPPresetsButton.button.enabled = state;
        }

        void CreateUI()
        {
            _presetTypePopup = CreateScrollablePopup(_presetTypeChooser);

            _setPresetPathButtonSelected = CreateButton("Set Preset Path");
            _setPresetPathButtonSelected.buttonColor = Color.red;
            _setPresetPathButtonSelected.button.onClick.AddListener(SetSelectedPath);

            _delaySlider = CreateSlider(_screenshotDelay, true);

            _processVAPPresetsButton = CreateButton("Process Preset Folder", true);
            _processVAPPresetsButton.buttonText.fontStyle = FontStyle.Bold;
            _processVAPPresetsButton.buttonColor = Color.red;
            _processVAPPresetsButton.button.onClick.AddListener(() =>
            {
                if (_savedPathDic[_selectedPresetType].val == "")
                {
                    _infoBox.Message = "Error: Please Select a valid folder.";
                }
                else
                {
                    _infoBox.Status = "Processing " + _selectedPresetType + " Presets";
                    StartCoroutine(ProcessVAPPresets());
                }
            });

            _processVAPPresetsButton.button.enabled = false;

            _retakeScreenShots = CreateButton("Retake Camera Positions");
            _retakeScreenShots.button.onClick.AddListener(() =>
            {
                StartCoroutine(StoreCameraPostions(true));
            });

            _infoBox.Status = "Set folder path to update Presets.";
        }

        IEnumerator StoreCameraPostions(bool force = false)
        {
            _needToSaveCams = false;
            Text help = SuperController.singleton.screenshotPreview.GetComponentsInChildren<Text>().Where(x => x.name == "HelpText").FirstOrDefault();
            _camPositions.OrigHelpText = help.text;
            _camPositions.OrigHelpColor = help.color;

            //use saved cam from scene or do we use a set new positions
            //if (camPositions.LoadCamPosition("VAMCAM") == null || force)
            //{
            //    camPositions.StoreCamPosition("VAMCAM");
            //}
            //record cam position vs head position
            if (_camPositions.CamPostionStatus("HeadPosition") == false || force)
            {
                _camPositions.StoreCamPosition("HeadPosition", _headTransform);
            }

            foreach (var presetType in _presets)
            {
                _presetName = presetType;
                if (presetType == _selectedPresetType)
                {
                    if (_camPositions.CamPostionStatus(_presetTypeID[presetType]) == false || force)
                    {
                        _camPositions.StoreCamPosition(_presetTypeID[presetType]);
                        _waitingOnScreenShot = true;
                        SuperController.singleton.DoSaveScreenshot(_cameraSavePath + _presetTypeID[presetType], RecordCamPos);
                        help.text = "Set Camera you want to use for " + presetType + " Preset";
                        help.color = _camHelpColorDic[presetType];

                        yield return new WaitWhile(() => _waitingOnScreenShot);
                        yield return new WaitForSecondsRealtime(1);

                    }
                }
            }

            help.text = _camPositions.OrigHelpText;
            help.color = _camPositions.OrigHelpColor;

            //FreeControllerV3 headController = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault();
            //SuperController.singleton.FocusOnController(headController, true, true);
        }

        void RecordCamPos(string presetType)
        {
            presetType = FileManagerSecure.GetFileName(presetType);
            _camPositions.StoreCamPositionFOV(presetType,SuperController.singleton.loResScreenShotCameraFOV);
            _waitingOnScreenShot = false;
        }

        void ScreenShotFinished(string text)
        {
            _waitingOnScreenShot = false;
        }

        void StopProcessing(bool forced = false)
        {
            if (forced && _isProcessing)
            {
                _infoBox.Message = "Processing stopped by user.";
            }
            else
            {
                _infoBox.Message = "Finished Proccessing all " + _selectedPresetType + " preset files.\nPresets processed = " + _presetCount;
                _infoBox.Status = "Ready to start Processing.";
                
            }
            EnableUIButtons(true);
            _stopRequested = false;
            _isProcessing = false;
            StopCoroutine("CheckForStopKeypress");
            //reset cam to default.
            //camPositions.LoadCamPosition("VAMCAM");

            //FreeControllerV3 headController = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault();
            //SuperController.singleton.FocusOnController(headController, true, true);
        }

        IEnumerator ProcessVAPPresets()
        {
            _presetCount = 0;
            _isProcessing = true;
            EnableUIButtons(false);
            StartCoroutine("CheckForStopKeypress", 2.0f);

            //for the first screenshot of session allow user to position camera for screenshots
            if (_needToSaveCams)
            {
                yield return StartCoroutine(StoreCameraPostions());
            }

            //check if camera positions needed are ready
            foreach (var presetType in _presets)
            {
                _presetName = presetType;
                if (presetType == _selectedPresetType && !_camPositions.CamPostionStatus(_presetTypeID[presetType]))
                {
                    yield return StartCoroutine(StoreCameraPostions());
                }
            }

            foreach (var presetType in _presets)
            {
                _presetName = presetType;
                if (presetType == _selectedPresetType)
                {
                    foreach (string presetname in _presetListDic[presetType])
                    {
                        if (_stopRequested)
                        {
                            StopProcessing(true);
                            yield break;
                        }
                        yield return StartCoroutine(ProcessVAPPreset(presetname));
                        yield return new WaitForEndOfFrame();
                    }

                }
            }

            yield return new WaitForEndOfFrame();

            StopProcessing();

            yield return null;
        }

        IEnumerator ProcessVAPPreset(string presetname)
        {
            _boolcontinue = false;
            _infoBox.Message = "Saving " + _selectedPresetType + " Preset " + presetname;
            yield return null;
            _presetCount++;
            _waitingOnScreenShot = true;

            containingAtom.GetStorableByID("EyelidControl").SetBoolParamValue("blinkEnabled", false);
            containingAtom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", false);
            containingAtom.GetStorableByID("TongueControl").SetBoolParamValue("tongueCollision", false);

            JSONStorable presetJS = containingAtom.GetStorableByID(_presetTypeID[_selectedPresetType]);
            _pm = presetJS.GetComponentInChildren<PresetManager>();
            _pm.postLoadEvent.AddListener(OnPresetLoaded);

            presetJS.GetStringJSONParam("presetName").val = presetname;

            yield return new WaitForEndOfFrame();
            yield return new WaitForEndOfFrame();
            //_pm.LoadPreset();
            presetJS.CallAction("LoadPreset");

            //JSONStorable presetJS;
            //presetJS = containingAtom.GetStorableByID(_presetTypeID[presetType]);
            //_pm = presetJS.GetComponentInChildren<PresetManager>();
            //_pm.postLoadEvent.AddListener(OnPresetLoaded);
            //_pm.presetName = "test";
            //presetJS.GetStringJSONParam("presetName").val = presetname;
            //presetJS.CallAction("LoadPreset");

            yield return new WaitUntil(() => _boolcontinue);

            //containingAtom.GetStorableByID(_presetTypeID[_selectedPresetType]).CallAction("StorePresetWithScreenshot");

            DAZCharacterRun dazCharacterRun = containingAtom.GetComponentInChildren<DAZCharacterSelector>().GetComponent<DAZCharacterRun>();
            DAZCharacter _dazCharacter = containingAtom.GetComponentInChildren<DAZCharacterSelector>().selectedCharacter;
            yield return new WaitUntil(() => _dazCharacter.ready);
            //wait for physics to finish
            yield return new WaitWhile(() => dazCharacterRun.IsSimulationResetting());
            //wait for clothing and any texture loads are finished
            yield return new WaitWhile(() => SuperController.singleton.loadingIcon.gameObject.activeSelf);
            //wait for physics to settle
            yield return new WaitForSeconds(_screenshotDelay.val);

            //_pm.postLoadEvent.RemoveListener(OnPresetLoaded);

            //yield return new WaitForSeconds(0.3f);

            yield return new WaitUntil(() => _pm.CheckPresetReadyForStore());

            _camPositions.UpdateHeadDeltas(_headTransform);
            _camPositions.LoadCamPosition(_presetTypeID[_selectedPresetType]);
            SuperController.singleton.loResScreenShotCameraFOV = _camPositions.LoadCamFOV(_presetTypeID[_selectedPresetType]);
            //_pm.StorePreset(true);
            //wait for camera positon update
            yield return new WaitForEndOfFrame();
            yield return new WaitForEndOfFrame();
            yield return new WaitForEndOfFrame();
            //presetJS.CallAction("StorePresetWithScreenshot");
            SuperController.singleton.DoSaveScreenshot(getSceenshotPathSimple(presetname), ScreenShotFinished);

            yield return new WaitForEndOfFrame();
            yield return new WaitForEndOfFrame();

            SuperController.singleton.SetLeftSelect();

            yield return new WaitWhile(() => _waitingOnScreenShot);
        }

        IEnumerator CheckForStopKeypress(float someParameter)
        {
            while (true)
            {
                if (Input.GetKeyDown(KeyCode.Space) || Input.GetKey(KeyCode.Space))
                {
                    _stopRequested = true;
                }

                yield return null;
            }
        }

        //public void Update()
        //{
        //    if (!_stopRequested && (Input.GetKeyDown(KeyCode.DownArrow) || Input.GetKeyDown(KeyCode.Space) || Input.GetKey(KeyCode.Space)))
        //    {
        //        _stopRequested = true;
        //    }
        //}

        private string getSceenshotPathSimple(string presetname)
        {
            string jpgPath = "";
            string[] strSplit = presetname.Split('/');
            
            for (int i = 0; i < strSplit.Length; i++)
            {
                if(i == strSplit.Length -1)
                {
                    jpgPath += "Preset_" + strSplit[i] + ".jpg";
                }
                else
                {
                    jpgPath += strSplit[i] + "/";
                }
            }
            return _atomPersonPath + _selectedPresetType + "/" + jpgPath;
        }

        private void OnPresetLoaded()
        {
            _boolcontinue = true;
        }

        void RemoveAllClothing()
        {
            var allGeometry = containingAtom.GetStorableByID("geometry").GetBoolParamNames();
            foreach (var item in allGeometry)
            {
                if (item.Contains("clothing:"))
                {
                    if (containingAtom.GetStorableByID("geometry").GetBoolJSONParam(item).val)
                    {
                        containingAtom.GetStorableByID("geometry").GetBoolJSONParam(item).val = false;
                    }
                }
            }
        }
    }

    class FilePaths
    {
        //the name and path to save the preset file. VAM will add 'Preset_' to this for saved VAP file and save in correct folder
        //ex /PresetMaker/MYNAME  becomes  "Custom/Atom/Person/Appearance/PresetMaker/Preset_MYNAME.vap
        public string StringParamValue { get { return presetSubPath + generatedName; } }
        string presetSubPath { get; set; }
        string generatedName { get; set; }
        public string fullVarName { get; private set; }
        public string varName { get; private set; }
        public string authorName { get; private set; }


        string appearancePath = "Custom/Atom/Person/Appearance";
        string clothingPath = "Custom/Atom/Person/Clothing";
        string skinPath = "Custom/Atom/Person/Skin";
        string hairPath = "Custom/Atom/Person/Hair";

        public string GetScreenShotPath(string presetType)
        {
            return PresetTypePath(presetType) + presetSubPath + "Preset_" + generatedName + ".jpg";
        }

        public string GetPresetPath(string presetType)
        {
            return PresetTypePath(presetType) + presetSubPath + "Preset_" + generatedName + ".vap";
        }

        string PresetTypePath(string presetType)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    return "Custom/Atom/Person/Appearance";

                case "ClothingPresets":
                    return "Custom/Atom/Person/Clothing";

                case "SkinPresets":
                    return "Custom/Atom/Person/Skin";

                case "HairPresets":
                    return "Custom/Atom/Person/Hair";
            }

            return null;
        }

        public void CreateFileName(string fullVarName, string name = "Person", bool isMale = false)
        {
            string sex = "Female";
            if (isMale)
                sex = "Male";

            this.fullVarName = fullVarName;
            string[] varNames = fullVarName.Split('.');

            authorName = varNames[0];
            varName = varNames[1];

            presetSubPath = string.Concat(new string[]
                {
                    "/",
                    "PresetMaker",

                    "/",
                    authorName,
                    "/",
                });

            if (name.ToLower().Contains("person"))
            {
                name = varName;
            }

            int count = 0;

            string baseName = name + "-" + sex;
            generatedName = baseName;
            string path = presetSubPath + "Preset_" + baseName + ".vap";
            while (FileExists(path))
            {
                count++;
                generatedName = baseName + count;
                path = presetSubPath + "Preset_" + generatedName + ".vap";
            }
        }

        bool FileExists(string path)
        {
            if (FileManagerSecure.FileExists(appearancePath + path) || FileManagerSecure.FileExists(clothingPath + path) || FileManagerSecure.FileExists(skinPath + path) || FileManagerSecure.FileExists(hairPath + path))
            {
                return true;

            }
            return false;
        }
    }

    class TransformData
    {
        public Vector3 localPosition { get; set; }
        public Vector3 position { get; private set; }
        public Quaternion localRotation { get; private set; }
        public Vector3 centerCamEulerAngles { get; private set; }
        public float playerHeightAdjust { get; set; }

        public Color OrigHelpColor { get; set; }
        public string OrigHelpText { get; set; }

        public TransformData()
        {
            localPosition = SuperController.singleton.navigationRig.transform.localPosition;
            localRotation = SuperController.singleton.navigationRig.transform.localRotation;
            position = SuperController.singleton.navigationRig.transform.position;

            playerHeightAdjust = SuperController.singleton.playerHeightAdjust;
            centerCamEulerAngles = SuperController.singleton.MonitorCenterCamera.transform.localEulerAngles;
        }

        public TransformData(Transform transform)
        {   //used for head delta and world position
            position = transform.position;
            localRotation = transform.rotation;
            playerHeightAdjust = SuperController.singleton.playerHeightAdjust;
            centerCamEulerAngles = SuperController.singleton.MonitorCenterCamera.transform.localEulerAngles;
        }

        public void RestoreTransform(Vector3 delta = new Vector3())
        {
            if (delta == null)
                delta = Vector3.zero;

            //Vector3 newPos = new Vector3(localPosition.x + delta.x, localPosition.y, localPosition.z + delta.z);
            SuperController.singleton.navigationRig.transform.localPosition = localPosition;
            SuperController.singleton.navigationRig.transform.localRotation = localRotation;
            SuperController.singleton.playerHeightAdjust = playerHeightAdjust + delta.y;
            SuperController.singleton.MonitorCenterCamera.transform.localEulerAngles = centerCamEulerAngles;
        }

        #region save-load
        //convert TransforData to JSON Object for storage
        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();

            if (localPosition == null || localRotation == null || centerCamEulerAngles == null)
                return null;

            try
            {
                jc["localPosition"]["x"].AsFloat = localPosition.x;
                jc["localPosition"]["y"].AsFloat = localPosition.y;
                jc["localPosition"]["z"].AsFloat = localPosition.z;

                jc["Rotation"]["x"].AsFloat = localRotation.x;
                jc["Rotation"]["y"].AsFloat = localRotation.y;
                jc["Rotation"]["z"].AsFloat = localRotation.z;
                jc["Rotation"]["w"].AsFloat = localRotation.w;

                jc["centerCamEulerAngles"]["x"].AsFloat = centerCamEulerAngles.x;
                jc["centerCamEulerAngles"]["y"].AsFloat = centerCamEulerAngles.y;
                jc["centerCamEulerAngles"]["z"].AsFloat = centerCamEulerAngles.z;

                jc["playerHeightAdjust"].AsFloat = playerHeightAdjust;
            }
            catch (Exception e)
            {
                SuperController.LogError("PresetMaker: Error Saving JSON\n" + e.Message);
            }
            return jc;
        }

        public void RestoreFromJSON(JSONClass jc)
        {
            try
            {
                localPosition = new Vector3(jc["localPosition"]["x"].AsFloat, jc["localPosition"]["y"].AsFloat, jc["localPosition"]["z"].AsFloat);
                localRotation = new Quaternion(jc["Rotation"]["x"].AsFloat, jc["Rotation"]["y"].AsFloat, jc["Rotation"]["z"].AsFloat, jc["Rotation"]["w"].AsFloat);

                centerCamEulerAngles = new Vector3(jc["centerCamEulerAngles"]["x"].AsFloat, jc["centerCamEulerAngles"]["y"].AsFloat, jc["centerCamEulerAngles"]["z"].AsFloat);
                playerHeightAdjust = jc["playerHeightAdjust"].AsFloat;
            }
            catch (Exception e)
            {
                SuperController.LogError("PresetMaker: Error loading JSON\n" + e.Message);
            }
        }
        #endregion
    }

    class CamPositions
    {
        public string OrigHelpText { get; set; }
        public Color OrigHelpColor { get; set; }

        Vector3 deltaP = new Vector3();
        Dictionary<string, TransformData> camDictionary = new Dictionary<string, TransformData>();
        Dictionary<string, float> camDictionaryFOV = new Dictionary<string, float>();

        public void UpdateHeadDeltas(Transform transform)
        {
            deltaP = transform.position - camDictionary["HeadPosition"].position;
        }

        public bool CamPostionStatus(string presetType)
        {
            if (camDictionary.ContainsKey(presetType))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        public TransformData LoadCamPosition(string presetType)
        {
            TransformData camData;
            if (camDictionary.TryGetValue(presetType, out camData))
            {
                camData.RestoreTransform(deltaP);
                return camData;
            }
            else
            {
                return null;
            }
        }

        public float LoadCamFOV(string presetType)
        {
            float camFOV;
            if (camDictionaryFOV.TryGetValue(presetType, out camFOV))
            {
                return camFOV;
            }
            else
            {
                return -1f;
            }
        }    

        public TransformData StoreCamPosition(string presetType, Transform transform)
        {
            return camDictionary[presetType] = new TransformData(transform);
        }

        public TransformData StoreCamPosition(string presetType)
        {
            return camDictionary[presetType] = new TransformData();
        }
        public TransformData StoreCamPositionFOV(string presetType, float intfov)
        {
            camDictionaryFOV[presetType] = intfov;
            return camDictionary[presetType] = new TransformData();
        }

    }

    public class InfoBox
    {
        JSONStorableString infoBox;
        int varCount = 0;
        int authorCount = 0;
        string status = "Waiting on Scan to Finishe before creating UI";
        string vapScan = "";
        int vapCount = 0;
        string message = "";
        string presetPath = "";

        public int VarCount { set { varCount = value; UpdateInfoBox(); } }
        public int AuthorCount { set { authorCount = value; UpdateInfoBox(); } }
        public string Status { set { status = value; UpdateInfoBox(); } }
        public string PresetPath { set { presetPath = value; UpdateInfoBox(); } }
        public string VapScan { set { vapScan = value; UpdateInfoBox(); } }
        public int VapCount { set { vapCount = value; UpdateInfoBox(); } }
        public string Message { set { message = value; UpdateInfoBox(); } }

        public InfoBox(JSONStorableString uiInfoBox)
        {
            infoBox = uiInfoBox;
        }

        public void UpdateInfoBox()
        {
            infoBox.SetVal(String.Format(formated, status, presetPath, vapCount, message));

        }

        string formated =
@"{0}
{1}
Presets found {2}

{3}
";


    }


}